@@ -87,7 +87,7 @@ struct ButWin_struct  {
 
   const char *name;
   bool  iconic, isIcon;
-  int  x,y,w,h, xOff, yOff, wRatio, hRatio;
+  int  x,y,w,h, xOff, yOff, minWRatio, minHRatio, maxWRatio, maxHRatio;
   int  minW, minH, maxW, maxH, wStep, hStep;
   int  logicalW, logicalH;  /* Same as w&h except for canvases. */
 
@@ -143,15 +143,14 @@ struct  ButEnv_struct  {
   int  (*sClear)(ButEnv *env);
   int  (*sNotify)(ButEnv *env, XSelectionEvent *xsnot);
   struct ButWin_struct  **winlist;
-  bool  keepAlive;
-  int  wllen, wlmax;
+  int  minWindows, wllen, wlmax;
   But  *butIn, *lockBut;
   struct ButWin_struct  *last_mwin;
   int  last_mx, last_my;
   ButEnv  *next;
   struct ButWrite_struct  write[BUTWRITE_MAXCHARS];
   uint  keyModifiers;
-  Time  eventTime;  /* But & kbd PRESS or RELEASE events. */
+  Time  eventTime;  /* But & kbd PRESS or RELEASE or MOVE events. */
   int  eventNum;   /* But & kbd PRESS events ONLY. */
 
   int  maxFd, maxGFds[3];
@@ -348,6 +347,7 @@ extern Snd  but_downSnd, but_upSnd;
 #define  butEnv_dpy(env)     ((env)->dpy)
 #define  butEnv_gc(env)      ((env)->gc)
 #define  butEnv_color(env, cnum)   ((env)->colors[cnum])
+#define  butEnv_isColor(env) ((env)->colorp)
 #define  butEnv_pixmap(env, cnum)  ((env)->colorPmaps[cnum])
 #define  butEnv_fontStr(env, fnum) ((env)->fonts[fnum])
 #define  butEnv_stdBw(env)   ((env)->stdButBw)
@@ -410,7 +410,6 @@ extern void  but_resize(But *but, int x, int y, int w, int h);
 #define  butEnv_context(env)       ((env)->context)
 extern void  but_setId(But *but, int id);
 extern void  butWin_setId(ButWin *win, int id);
-#define  butEnv_setKeepAlive(env, k)  ((env)->keepAlive = (k))
 
 #define  BUT_READFILE   0
 #define  BUT_WRITEFILE  1
@@ -427,6 +426,10 @@ extern ButWin  *butWin_iCreate(void *packet, ButEnv *env,
 			       ButWinFunc *resize,
 			       ButWinFunc *iResize,
 			       ButWinFunc *destroy);
+/*
+ * DO NOT ACTIVATE a root window.
+ */
+extern ButWin  *butWin_rootCreate(void *packet, ButEnv *env);
 #define  butWin_create(p,e,n,w,h,un,ma,re,de)  \
   butWin_iCreate(p,e,n,w,h,NULL,FALSE,0,0,un,ma,re,NULL,de)
 extern void  butWin_setGeom(ButWin *win, const char *geometry);
@@ -437,10 +440,14 @@ extern void  butWin_setMinH(ButWin *win, int minH);
 extern void  butWin_setMaxW(ButWin *win, int maxW);
 extern void  butWin_setMaxH(ButWin *win, int maxH);
 extern void  butWin_setWHRatio(ButWin *win, int w, int h);
+extern void  butWin_setWHRatios(ButWin *win,
+				int minW, int minH, int maxW, int maxH);
 extern void  butWin_setWStep(ButWin *win, int wStep);
 extern void  butWin_setHStep(ButWin *win, int hStep);
 #define  butWin_getMinW(w)   ((w)->minW)
 #define  butWin_getMinH(w)   ((w)->minH)
+#define  butWin_getMaxW(w)   ((w)->maxW)
+#define  butWin_getMaxH(w)   ((w)->maxH)
 #define  butWin_getWStep(w)  ((w)->wStep)
 #define  butWin_getHStep(w)  ((w)->hStep)
 extern void  butWin_activate(ButWin *win);
