@@ -1,6 +1,6 @@
 /*
  * wmslib/src/but/menu.c, part of wmslib (Library functions)
- * Copyright (C) 1994-1995 William Shubert.
+ * Copyright (C) 1994-1997 William Shubert.
  * See "configure.h.in" for more copyright information.
  *
  * Known problems:
@@ -31,10 +31,14 @@
 #include <but/menu.h>
 #include <but/box.h>
 #include <but/menu_snd.h>
+#include <wms/str.h>
 
 
+/**********************************************************************
+ * Data structures
+ **********************************************************************/
 typedef struct Ol_struct  {  /* Option List */
-  const char *text;
+  Str  text;
   int  x,y,w,h;
   uint  flags;
 } Ol;
@@ -53,9 +57,10 @@ typedef struct menu_struct  {
   ButOut  (*callback)(But *but, int value);
 } Menu;
 
-
+/**********************************************************************
+ * Forward Declarations
+ **********************************************************************/
 static int  menu_width(But *but);
-
 static ButOut  smmove(But *but, int x, int y);
 static ButOut  smleave(But *but);
 static ButOut  smpress(But *but, int butnum, int x, int y);
@@ -71,22 +76,29 @@ static void  ldraw(But *but, int x,int y, int w,int h);
 static ButOut  ldestroy(But *but);
 static void  lnewflags(But *but, uint flags);
 static ButOut  netMsg(But *but, void *msg, int msgLen);
-
 static But  *create(ButOut (*func)(But *but, int value), void *packet,
 		    ButWin *win, int layer, int toplayer, int flags,
 		    const char *title, const char *optlist[], int cur_opt,
 		    bool up);
 
-static ButAction  saction = {
+
+/**********************************************************************
+ * Globals
+ **********************************************************************/
+static const ButAction  saction = {
   smmove, smleave, smpress, NULL,
   NULL, NULL, sdraw, sdestroy, snewflags, netMsg};
 
-static ButAction  laction = {
+static const ButAction  laction = {
   lmmove, lmleave, lmpress, lmrelease,
   NULL, NULL, ldraw, ldestroy, lnewflags, NULL};
 
 char  butMenu_dummy = '\0';
 
+
+/**********************************************************************
+ * Functions
+ **********************************************************************/
 But  *butMenu_upCreate(ButOut (*func)(But *but, int value), void *packet,
 		       ButWin *win, int layer, int toplayer, int flags,
 		       const char *title, const char *optlist[],
@@ -131,8 +143,16 @@ static But  *create(ButOut (*callback)(But *but, int value), void *packet,
   m->options = wms_malloc((i+1) * sizeof(Ol));
   i = 0;
   do  {
-    m->options[i].text = optlist[i];
-    m->options[i].flags = 0;
+    if (optlist[i] == BUTMENU_OLBREAK)  {
+      str_init(&m->options[i].text);
+      m->options[i].flags = BUTMENU_BREAK;
+    } else if (optlist[i] == BUTMENU_OLEND)  {
+      str_init(&m->options[i].text);
+      m->options[i].flags = BUTMENU_END;
+    } else  {
+      str_initChars(&m->options[i].text, optlist[i]);
+      m->options[i].flags = 0;
+    }
   } while (optlist[i++] != BUTMENU_OLEND);
   m->llayer = toplayer;
   m->tval = 0;
@@ -165,8 +185,16 @@ void  butMenu_setText(But *but, const char *title, const char *optlist[],
   m->options = wms_malloc((i+1) * sizeof(Ol));
   i = 0;
   do  {
-    m->options[i].text = optlist[i];
-    m->options[i].flags = 0;
+    if (optlist[i] == BUTMENU_OLBREAK)  {
+      str_init(&m->options[i].text);
+      m->options[i].flags = BUTMENU_BREAK;
+    } else if (optlist[i] == BUTMENU_OLEND)  {
+      str_init(&m->options[i].text);
+      m->options[i].flags = BUTMENU_END;
+    } else  {
+      str_initChars(&m->options[i].text, optlist[i]);
+      m->options[i].flags = 0;
+    }
   } while (optlist[i++] != BUTMENU_OLEND);
   if (newopt != BUT_NOCHANGE)
     m->cval = newopt;
@@ -275,7 +303,7 @@ static void  sdraw(But *but, int x,int y, int w,int h)  {
       bbw = 1;
     sbw = (bbw + 1) / 2;
     bwdiff = bbw - sbw;
-    but_drawBox(win, but->x,but->y, but->w,but->h, 0, bwdiff,
+    but_drawBox(win, but->x, but->y, but->w, but->h, 0, bwdiff,
 		BUT_SLEFT|BUT_SRIGHT, BUT_LIT, BUT_SHAD, None, None);
     but_drawCt(win, flags, m->fgpic, m->bgpic, BUT_PBG,
 	       but->x+bwdiff, but->y+bwdiff,
@@ -284,16 +312,21 @@ static void  sdraw(But *but, int x,int y, int w,int h)  {
     but_drawCt(win, flags, BUT_FG,BUT_BG,BUT_PBG,
 	       but->x+bwdiff, but->y + but->h/2,
 	       but->w-2*bwdiff, but->h - but->h/2 - bwdiff,
-	       sbw, m->options[m->cval].text, BUT_SLEFT|BUT_SRIGHT, 0);
+	       sbw, str_chars(&m->options[m->cval].text),
+	       BUT_SLEFT|BUT_SRIGHT, 0);
   }
 }
 
 
 static ButOut  sdestroy(But *but)  {
   Menu  *m = but->iPacket;
+  int  i;
 
   if (m->child != NULL)
     but_destroy(m->child);
+  for (i = 0;  i <= m->nvals;  ++i)  {
+    str_deinit(&m->options[i].text);
+  }
   wms_free(m->options);
   wms_free(but->iPacket);
   return(0);
@@ -349,7 +382,7 @@ static But  *lcreate(But *smenu, bool local)  {
   w = but->w - 2*env->stdButBw;
   h = env->font0h + env->stdButBw;
   for (i = 0;  i < m->nvals;  ++i)  {
-    if (m->options[i].text == BUTMENU_OLBREAK)  {
+    if (m->options[i].flags & BUTMENU_BREAK)  {
       y += (env->stdButBw & ~1) + env->stdButBw;
     } else  {
       m->options[i].x = x;
@@ -362,9 +395,16 @@ static But  *lcreate(But *smenu, bool local)  {
   m->tval = m->cval;
   m->last_w = 0;
   m->clickOpen = FALSE;
-  if (local)
-    XWarpPointer(env->dpy, None, win->win, 0,0,0,0,
-		 but->x + but->w/2, m->options[m->cval].y - butbw + sh/2);
+  if (local)  {
+    x = but->x + but->w/2;
+    y = m->options[m->cval].y - butbw + sh/2;
+    while (win->parent != NULL)  {
+      x += win->xOff;
+      y += win->yOff;
+      win = win->parent;
+    }
+    XWarpPointer(env->dpy, None, win->win, 0,0,0,0, x, y);
+  }
   m->skip_clicks = TRUE;
   but_init(but);
   snd_play(&butMenu_openSnd);
@@ -387,7 +427,7 @@ static ButOut  lmmove(But *but, int x, int y)  {
   if (but->flags & BUT_PRESSED)
     result |= BUTOUT_CAUGHT;
   for (msel = 0;  msel < m->nvals;  ++msel)  {
-    if (m->options[msel].text != BUTMENU_OLBREAK)  {
+    if (!(m->options[msel].flags & BUTMENU_BREAK))  {
       m_y = m->options[msel].y - butbw;
       if ((y > m_y + butbw / 2) &&
 	  (y <= m_y + butbw / 2 + 2*butbw + env->font0h))
@@ -529,11 +569,10 @@ static void  ldraw(But *but, int x,int y, int w,int h)  {
   /* Fill in the background for the options. */
   butEnv_setXFg(env, BUT_BG);
   XFillRectangle(env->dpy, win->win, env->gc,
-		 but->x + butbw, new_y,
+		 but->x + butbw - win->xOff, new_y - win->yOff,
 		 but->w - butbw * 2, but->y + but->h - new_y - butbw);
   for (i = 0;  i <= m->nvals;  ++i)  {
-    if ((m->options[i].text == BUTMENU_OLBREAK) ||
-	(m->options[i].text == BUTMENU_OLEND))  {
+    if (m->options[i].flags & (BUTMENU_BREAK | BUTMENU_END))  {
       /* Draw the inner box around a group of options. */
       tx = but->x + (butbw+1)/2;
       ty = old_y;
@@ -559,7 +598,8 @@ static void  ldraw(But *but, int x,int y, int w,int h)  {
 		      None, None);
 	  butEnv_setXFg(env, BUT_HIBG);
 	  XFillRectangle(env->dpy, win->win, env->gc,
-			 but->x + 2*butbw, new_y + butbw,
+			 but->x + 2*butbw - win->xOff,
+			 new_y + butbw - win->yOff,
 			 but->w - 4*butbw, env->font0h + butbw);
 	}
       }
@@ -571,7 +611,7 @@ static void  ldraw(But *but, int x,int y, int w,int h)  {
 	} else
 	  butEnv_setXFg(env, BUT_FG);
 	butWin_write(win, but->x + m->xoff, m->options[i].y + butbw/2,
-		     m->options[i].text, 0);
+		     str_chars(&m->options[i].text), 0);
 	if (m->options[i].flags & BUTMENU_DISABLED)  {
 	  butEnv_stdFill(env);
 	}
@@ -587,10 +627,22 @@ static ButOut  ldestroy(But *but)  {
   ButEnv  *env = win->env;
   Menu  *m = but->iPacket;
   But  *sbut = m->parent;
+  int  x, y;
+  int  minX, minY;
 
+  minX = but->x;
+  minY = but->y;
+  x = sbut->x + sbut->w/2;
+  y = sbut->y + sbut->h/2;
+  while (win->parent != NULL)  {
+    minX += win->xOff;
+    x += win->xOff;
+    minY += win->yOff;
+    y += win->yOff;
+    win = win->parent;
+  }
   XWarpPointer(env->dpy, win->win, win->win,
-	       but->x, but->y, but->w, but->h,
-	       sbut->x + sbut->w/2, sbut->y + sbut->h/2);
+	       minX, minY, but->w, but->h, x, y);
   m->child = NULL;
   return(0);
 }
@@ -611,9 +663,9 @@ static int  menu_width(But *but)  {
   int  i, maxw, tw;
   
   maxw = butEnv_textWidth(env, m->title, 0);
-  for (i = 0;  m->options[i].text != BUTMENU_OLEND;  ++i)  {
-    if (m->options[i].text != BUTMENU_OLBREAK)  {
-      tw = butEnv_textWidth(env, m->options[i].text, 0);
+  for (i = 0;  !(m->options[i].flags & BUTMENU_END);  ++i)  {
+    if (!(m->options[i].flags & BUTMENU_BREAK))  {
+      tw = butEnv_textWidth(env, str_chars(&m->options[i].text), 0);
       if (tw > maxw)
 	maxw = tw;
     }
@@ -652,4 +704,16 @@ static ButOut  netMsg(But *but, void *msg, int msgLen)  {
 }
 
 
+void  butMenu_setOptionName(But *but, const char *new, int entryNum)  {
+  Menu  *m;
+
+  assert(but->action == &saction);
+  m = but->iPacket;
+  assert(entryNum < m->nvals);
+  str_copyChars(&m->options[entryNum].text, new);
+  if (entryNum == m->cval)
+    but_draw(but);
+}
+
+
 #endif
